#! /bin/perl
#
# jhttp.pl	--- send http requests, janl's 12" mix.
#
# This implements http/1.0 requests.  We'll have problems with http/0.9
#
# This is a rewrite of http.pl by Oscar Nierstrasz; I copied the code he he
# copied from the camel book.  Some functions written by Gorm Haug Eriksen
# (gorm@usit.uio.no) has been used as is.
#
# NB: If this package interests you, you should probably have a look at
# Roy Fielding's libwww-perl packages: 
# http://www.ics.uci.edu/WebSoft/libwww-perl/
#
# gorm :
# http'get_last_modified  return the last modified stamp on a file in
#                         the right format for use with http
# http'bit2rfc850  convert from 32 bit timestamp to rfc850
#
# janl:
# &http'query: Send a http query.  A completely general function to send a
#   http query.  Will extract header values, http response code and, optionaly,
#   convert text files to local linefeed format.
# &http'encode_url: Encode a url according to rfc1738/December 1994
#
# Variables to examine after a query
# $http'document: The document returned by the query, if any.
# $http'result: The numerical http result code
# $http'restext: The english(?) http result
# $http'header: The http header returned.
# $http'plaintexthtml: 1 if this doc is non-content-encoded text/html
#	(as opposed to content-encoding: compressed content-type: text/html
#       which needs decompression before we can inspect the html) 
#	The test is somewhat longwinded so I do it just once here.
# %http'headval: Associative array of header values
# $http'xfbytes: Transfered bytes, cumulative.
#
# Variables that change http's behaviour/requests:
# $http'agent: User agent, default is basename of $0
# $http'from: Request is from, default is user@host
# $http'version: The http version to use, only 1.0 is known
# $http'timeout: How long to wait for new data to arrive, default is 600sec
# $http'buflen: Network read buffer size, default is 4096.  It might give a
#	speedup to tune this for specific servers' so it matches their send
#	size.
# $http'debug: 1 debuging output, 2, more, 3 queries and replies
# $http'convert: Convert text/* documents to local newline convention?
#	The default is to do it.
# $http'proxyserver: The name of the proxyserver to use
# $http'proxyport: The port of the proxyserver to use. 0 if no proxyserver.
#
# Missing things:
# - A function that retries the query if it times out, folows redirects and
#   other useful things.
# - the accept variable has been removed: a Accept: is not part of the http
#   spec.
#
# Oscar Nierstrasz oscar@cui.unige.ch
# Gorm Haug Eriksen gorm@usit.uio.no
# Nicolai Langfeldt janl@ifi.uio.no

package http;

# This should be in /local/lib/perl. If it's not there, complain to your 
# system admin!
require "sys/socket.ph";

# Find out some things
$sockaddr = 'S n a4 x8';
chop($thishost = `hostname`);
($name, $aliases, $proto) = getprotobyname("tcp");

($name, $aliases, $type, $len, $thisaddr) = gethostbyname($thishost);
chop($user = `whoami`);
$from   = "$user@$name";

$thissock = pack($sockaddr, &AF_INET, 0, $thisaddr);

$nl = "\r\n";
# Default values, change by assignment from user
$agent  = $0; $agent =~ s~.*/~~; # Basename
$version= "1.0";
$timeout= 600;			# Timeout while waiting for data/connection
$buflen = 4096;			# recv buffer length
$debug  = 0;			# Debuging output?
$convert= 1;			# Convert newlines of text docs to local format
$proxyserver='';		# Proxy server.
$proxyport=0;			# Proxy server port. 0 if no proxy.
$xfbytes=0;			# 0 bytes transfered thus far.

# Query opcodes
$GET = 1;			# GET query. Arg: host,port,path
$HEAD = 2;			# HEAD query. Arg: host,port,path
$GETURL = 3;			# GET query. Arg: url
$HEADURL = 4;			# HEAD query. Arg: url
# Modify query thus:
$IFMOD   = 101;			# If-modified after: Arg: HTTP-date-str
$IFMODF  = 102;			# If-modified after file: Arg: local-file-name
$AUTHORIZ= 103;			# Authorization. Arg: credentials

sub query {
    # Build and send a HTTP query.  And also receive response - janl 95/09/18

    # We do next to no argument type checking btw.

    local($host,$port,$request,$lastmod,$query,$method);

    $result=100;

    if ($version ne '1.0') {
	warn "Unknown HTTP version $version, no request sent\n";
	return 0;
    }


    $restext='internal error';
    $query='';

    # Find out what to ask for

    while (defined($arg=shift)) {
	if ($arg == $GET) {
	    $host=shift;
	    $port=shift;
	    $request=&encode_url(shift);
	    if ($proxyport) {
		$query.="GET http://".$host.":".$port.$request;
	    } else {
		$query.="GET ".$request;
	    }
	    $query.=" HTTP/".$version.$nl;
	} elsif ($arg == $HEAD) {
	    $host=shift;
	    $port=shift;
	    $request=&encode_url(shift);
	    if ($proxyport) {
		$query.="HEAD http://".$host.":".$port.$request;
	    } else {
		$query.="HEAD ".$request;
	    }
	    $query.=" HTTP/".$version.$nl;
	} elsif ($arg == $GETURL) {
	    if ($proxyport) {
		$query.="GET ".&encode_url(shift);
	    } else {
		($method,$host,$port,$request)=&parse_url(shift);
		$query.="GET ".&encode_url($request);
	    }
	    $query.=" HTTP/".$version.$nl;
	} elsif ($arg == $HEADURL) {
	    if ($proxyport) {
		$query.="HEAD ".&encode_url(shift);
	    } else {
		($method,$host,$port,$request)=&parse_url(shift);
		$query.="HEAD ".&encode_url($request);
	    }
	    $query.=" HTTP/".$version.$nl;
	} elsif ($arg == $IFMOD) {
	    $query.="If-Modified-Since: ".(shift).$nl;
	} elsif ($arg == $IFMODF) {
	    $query.="If-Modified-Since: ".&last_modified(shift).$nl;
	} elsif ($arg == $AUTHORIZ) {
	    $query.="Authorization: ".(shift).$nl;
	} else {
	    warn "Unknown http query opcode: $arg\n";
	}
	# Insert the last parts of the query:
    }

    $query.="User-Agent: ".$agent.$nl."From: ".$from.$nl.$nl;

    print STDERR "QUERY:\n",$query,"\n---\n" if $debug>=3;

    # If we're using proxy then set up the below code.
    if ($proxyport) {
	$host=$proxyserver;
	$port=$proxyport;
    }

    # Find out who to ask

    if (defined($address{$host})) {
	# Cache lookup! Revuloion!
	$thataddr=$address{$host};
    } else {
	# Cache miss, get and remember.
	($fqdn, $aliases, $type, $len, $thataddr) = gethostbyname($host);
	
	print STDERR "Lookup of $host:\nFQDN: $fqdn\nAliases: $aliases\n"
	    if $debug;

	$address{$host}=$thataddr;
	$address{$fqdn}=$thataddr if $fqdn ne $host;
    }

    $thatsock = pack($sockaddr, &AF_INET, $port, $thataddr);
    $check=getprotobyname("tcp");
    socket(FS, &AF_INET, &SOCK_STREAM, $proto) || return undef;
    warn "Got my socks on\n" if $debug;
    bind(FS, $thissock) || return undef;
    warn "Bound and waiting for it\n" if $debug;
    connect(FS, $thatsock) || return undef;
    warn "Connected\n" if $debug;

    return &resetsign if ($chime);

    # Arrange timeout
    $savALRM = $SIG{'ALRM'};
    $chime=0;			# There has been no alarm yet
    $SIG{'ALRM'}="http'timeout";
    alarm($timeout);

    send(FS,$query,0) || return &resetsign;
    return &resetsign if $chime;

    $answer='';
    $inp=' 'x$buflen;
    # Retrive HTTP response.  Why do I use recv and not <FS>? Because then
    # the timeout can work correctly!
    while (1) {
	recv(FS,$inp,$buflen,0);
	# If there has been a timeout, then we quit now.  The recv man page
        # does not seem to allow recv to return the bytes received up to
        # the timeout.
	if ($chime) {
	    warn "Timeout while waiting for HTTP reply to finish\n";
	    $!=0;
	    return &resetsign;
	}
	alarm($timeout);
	$answer.=$inp;
	last unless length($inp);
    }

    $xfbytes += length($answer);

    # Pick out http result code
    # Split the answer at the first two consecutive CRLFs.  'cept that
    # some servers just send two LFs. I hope none send two CRs
    ($header,$document) = split(/\r?\n\r?\n/,$answer,2);

    undef $answer;		# Save some memory

    warn "Header:\n$header\n---\n" if $debug>=3;

    ($result,$restext) = $header =~ m~^HTTP/\d\.\d (\d\d\d) (.*)~;

    warn "Result: $result, Text: $restext\n" if $debug>=2;

    # Shave off http result code from the header
    $header =~ s~^.*\n~~;

    local($*)=1;
    # Summary of the http spec on headers (with my comments):
    # - Each header line ends in CRLF (or just LF, or maybe even just CR,
    #   anyways, it's easier if all is LF).
    $header =~ s/\r/\n/g unless $header =~ s/\r\n/\n/g;
    # - If a line starts with space then it's a continuation of the previous
    #   line (these I fold into one line).
    $header =~ s/\n\s/ /;
    # - The header field names are case insensitive (so I convert them to
    #   lowercase)
    # - A field may appear twice, that is equivalent to listing the values
    #   in a comma separated list (so I fold them into a comma separated list)n
    # - The field name and the field value are separated by ': '

    # Clobber twice, gawd perl 4.036 is buggy!
    %headval=(); undef %headval;

    foreach $hdln (split(/\r?\n/,$header)) {
	($key,$value)=split(': ',$hdln,2);
	$key="\L$key";
	print STDERR "K: '$key', V: '$value'\n" if $debug>=2;
	if (defined($headval{$key})) {
	    $headval{$key}.=", ".$value;
	} else {
	    $headval{$key}=$value;
	}
    }

    # Paranoia check. Make it conditional? Nah, it making noise has mostly
    # been a indication of programming error on my part.
    if (defined($headval{'content-length'}) &&
	$headval{'content-length'} != length($document)) {
	$wanted=$headval{'content-lengh'};
	$got=length($document);
	warn "Content length ($got) differs from length declared in header ($wanted)\n";
    }

    # Check if the document is a non-encoded html document. The contents
    # could be (x-)?compress or (x-)gzip coded (compressed in other words).
    # html tagene uten } gj|re noe mer.
    $plaintexthtml=defined($http'headval{'content-type'}) &&
		$http'headval{'content-type'} eq 'text/html' &&
		!defined($http'headval{'content-encoding'});

    # If this is a non-encoded text file and we're supposed to convert
    # foreign newlines then we do it.
    if ($convert &&
	defined($http'headval{'content-type'}) &&
	$http'headval{'content-type'} =~ m~text/.*~ &&
	!defined($http'headval{'content-encoding'})) {
	# Change non unix newlines to unix newlines. bare \r is known from 
    	# macintosh (they hadta be different didn't they?), \r\n is known
	# as 'network format' and from numerous systems, among them ms-dos.
	$document =~ s~\r~\n~g unless $document =~ s~\r\n~\n~g;
	warn "Newlines converted(?)\n" if $debug;
    }

    &resetsign;
    1;
}


# Utilities for query

sub parse_url {
    # Parse a absolute url.
    local($url)=@_;

    local($method,$server,$port,$path) =
	$url =~ m~^([-.+A-Za-z0-9]+)://([^:/]+)(:\d+)?(.*)$~;

    if (defined($port) && $port) {
	$port =~ s/^://;
    } else {
	$port=80;
    }

    print STDERR "Method: $method\nServer: $server\nPort: $port\nPath:$path\n"
	if $debug>=2;

    return ($method,$server,$port,$path);
}



sub last_modified {
# will return the last modified time for a local file
# this procedure are for mirroring. The return will be in the
# rfc 850 format, and the timezone will be GMT - gorm

    local($file) = @_;
    local(@tmp) = stat($file); # file doesn't exist ok to fetch
    # now we got the last modified in a 32 bit integer.
    # time to convert it and return
    return &bit2rfc850($tmp[9]);
}

    
sub bit2rfc850 {
# this procedure will convert a 32bit timefield to regular
# rfc850 GMT format. this is implemented in this package because
# this format is the format used by http - gorm
# IN  : 32bit timesign
# OUT : http formated timestamp
    local($timebit) = @_;
    local(@DoW) = ('Sunday','Monday','Tuesday','Wedensday','Thursday','Friday','Saturday');
    local(@MoY) = ('Jan','Feb','Mar','Apr','May','Jun',
	    'Jul','Aug','Sep','Oct','Nov','Dec'); 

    local($time) = @_;
    local($[) = 0;
    local($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst);
    ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) = 
	gmtime($timebit);	# uses GMT time
# hack to fix the real time #########################################
    @tmplist = ($mday,$hour,$min,$sec);                             #
    for ($c=0;$c <= $#tmplist;$c++) {                               #
	$tmplist[$c] = "0$tmplist[$c]" if $tmplist[$c] < 10;        #
    }				                                    #
    ($mday,$hour,$min,$sec) = @tmplist;                             #
#####################################################################
# returning the right http format	
    sprintf("%s, %s-%s-%s %s:%s:%s GMT", 
	    $DoW[$wday], $mday, $MoY[$mon], $year, $hour, $min, $sec)
}


sub timeout {
    # Flag that we timed out - janl 95/09/18
    $chime=1;			# When this is 1 then the alarm has gone off
    warn "Timeout!\n";
}


sub resetsign {
    # Reset signal handlers - janl 95/09/18
    return 0 if !defined($savALRM);
    $SIG{'ALRM'}=$savALRM;
    return 0;
}

# RFC1738 specifies a wide range of characters to encode, we encode
# most of them, but not all, so as not to complicate things
# unnecesarily (we need to pick the url into its parts to do it
# Right)

sub encode_url {
    $url=shift;

    $url =~ 
	s/([\x00-\x20"#%~;<>\x7F-\xFF])/sprintf("%%%02lx",unpack('C',$1))/ge;

    return $url;
}

1;
